#include "EnergyEvaluation.h"
#include "parameters.h"
#include <cstdint>
#include <cmath>
#include "Utilities.h"
#include <thread>

using namespace cv;
using namespace std;
using namespace Utilities;

//Forward declarations of helper functions
static double transform_sigma_EmpDim(double sigma, double epsilon, double delta, double C1, double C2, double threshold);
static double transform_sigma_NucNorm(double sigma, double threshold);
static Mat build_M(Mat& xbar, Mat& History);
static void center_M(cv::Mat& M);

double compute_P(Mat& xbar, Mat& History, MultiTrackerParameters *TrackerParams) {
	switch(TrackerParams->penaltyForm) {
		case MultiTrackerParameters::penaltyType::None:                  return(0.0);
		case MultiTrackerParameters::penaltyType::EmpiricalDimension:    return(compute_P_EmpDim (xbar, History, TrackerParams));
		case MultiTrackerParameters::penaltyType::NuclearNorm:           return(compute_P_NucNorm(xbar, History, TrackerParams));
		case MultiTrackerParameters::penaltyType::ExplicitFactorization: return(compute_P_ExpFact(xbar, History, TrackerParams));
		default: fprintf(stderr,"Unimplemented penalty function for multi-feature tracker.\n");
		         return(0.0);
	}
}

void compute_P(Mat& xbar, Mat& History, double& P, Mat& DP, MultiTrackerParameters *TrackerParams) {
	switch(TrackerParams->penaltyForm) {
		case MultiTrackerParameters::penaltyType::None:
			//Generate reasonable default values for P and DP
			fprintf(stderr,"compute_P with jacobian evaluation should not be called with penaltyType=None.");
			break;
		case MultiTrackerParameters::penaltyType::EmpiricalDimension:    compute_P_EmpDim (xbar, History, P, DP, TrackerParams); break;
		case MultiTrackerParameters::penaltyType::NuclearNorm:           compute_P_NucNorm(xbar, History, P, DP, TrackerParams); break;
		case MultiTrackerParameters::penaltyType::ExplicitFactorization: compute_P_ExpFact(xbar, History, P, DP, TrackerParams); break;
		default: fprintf(stderr,"Unimplemented penalty function for multi-feature tracker.\n");
	}
}

//Compute P - Empirical Dimension Form
double compute_P_EmpDim(Mat& xbar, Mat& History, MultiTrackerParameters *TrackerParams) {
	double epsilon = TrackerParams->epsilon;
	double delta = epsilon/(1-epsilon);
	
	Mat M = build_M(xbar, History);
	if (TrackerParams->centerHistoryMatrix) center_M(M);
	
	//Compute singular values of M
	Mat sigma;
	SVD SVDObject;
	SVDObject.compute(M, sigma, SVD::NO_UV | SVD::MODIFY_A);
	
	//Evaluate Empirical Dimension
	Mat temp;
	cv::pow(sigma, epsilon, temp);
	double epsilonNorm = pow((double)(sum(temp).val[0]), 1.0/epsilon);
	
	cv::pow(sigma, delta, temp);
	double deltaNorm = pow((double)(sum(temp).val[0]), 1.0/delta);
	
	//Evaluate P
	return((TrackerParams->penaltyCoefficient)*(epsilonNorm/deltaNorm));
}


//Compute P and its Jacobian, DP - Empirical Dimension Form
void compute_P_EmpDim(Mat& xbar, Mat& History, double& P, Mat& DP, MultiTrackerParameters *TrackerParams) {
	double sigmaThreshold = 0.05;
	int F = (xbar.rows)/2;
	
	double epsilon = TrackerParams->epsilon;
	double delta = epsilon/(1-epsilon);
	
	Mat M = build_M(xbar, History);
	if (TrackerParams->centerHistoryMatrix) center_M(M);
	
	//Compute singular value decomposition of M
	Mat sigma, U, Vt;
	SVD SVDObject;
	SVDObject.compute(M, sigma, U, Vt, SVD::MODIFY_A);
	
	//Evaluate Empirical Dimension
	Mat temp;
	cv::pow(sigma, epsilon, temp);
	double epsilonNorm = pow((double)(sum(temp).val[0]), 1.0/epsilon);
	
	cv::pow(sigma, delta, temp);
	double deltaNorm = pow((double)(sum(temp).val[0]), 1.0/delta);
	
	//Evaluate P
	P = (TrackerParams->penaltyCoefficient)*(epsilonNorm/deltaNorm);
	
	//Evaluate DP
	double C1 = pow(epsilonNorm, 1.0-epsilon)/deltaNorm;
	double C2 = epsilonNorm * pow(deltaNorm, -1.0-delta);
	
	for (int n=0; n<sigma.rows; n++)
		sigma.at<double>(n) = transform_sigma_EmpDim(sigma.at<double>(n), epsilon, delta, C1, C2, sigmaThreshold);
	Mat Sigma = Mat::diag(sigma);
	
	Mat dd_e_dM = U*Sigma*Vt;
	Mat A = dd_e_dM(Rect(0,History.rows,F,2));
	Mat C;
	
	if (TrackerParams->centerHistoryMatrix) {
		Mat B;
		reduce(A, B, 1, CV_REDUCE_SUM, CV_64FC1);
     	C = (TrackerParams->penaltyCoefficient)*(A - (1.0/((double)F))*repeat(B, 1, F));
     }
     else {
     	C = (TrackerParams->penaltyCoefficient)*A;
     }
     
     //Copy elements of C into DP in correct order
     DP = Mat(1, 2*F, CV_64FC1);
     int x = 0;
     int y = 0;
     for (int n = 0; n<2*F; n++) {
     	DP.at<double>(n) = C.at<double>(x, y);
     	x++;
     	if (x == 2) {
     		x = 0;
     		y++;
     	}
     }
}

//Compute P - Nuclear Norm Form
double compute_P_NucNorm(cv::Mat& xbar, cv::Mat& History, MultiTrackerParameters *TrackerParams) {
	Mat M = build_M(xbar, History);
	if (TrackerParams->centerHistoryMatrix) center_M(M);
	
	//Compute singular values of M
	Mat sigma;
	SVD SVDObject;
	SVDObject.compute(M, sigma, SVD::NO_UV | SVD::MODIFY_A);
	
	return((TrackerParams->penaltyCoefficient)*((double)(sum(sigma).val[0])));
}

//Compute P and its Jacobian, DP - Nuclear Norm Form
void compute_P_NucNorm(cv::Mat& xbar, cv::Mat& History, double& P, cv::Mat& DP, MultiTrackerParameters *TrackerParams) {
	double sigmaThreshold = 0.05;
	int F = (xbar.rows)/2;
	
	Mat M = build_M(xbar, History);
	if (TrackerParams->centerHistoryMatrix) center_M(M);
	
	//Compute singular value decomposition of M
	Mat sigma, U, Vt;
	SVD SVDObject;
	SVDObject.compute(M, sigma, U, Vt, SVD::MODIFY_A);
	
	//Evaluate P
	P = (TrackerParams->penaltyCoefficient)*((double)(sum(sigma).val[0]));
	
	//Evaluate DP
	for (int n=0; n<sigma.rows; n++)
		sigma.at<double>(n) = transform_sigma_NucNorm(sigma.at<double>(n), sigmaThreshold);
	Mat Sigma = Mat::diag(sigma);
	
	Mat dNucNorm_dM = U*Sigma*Vt;
	
	
	Mat A = dNucNorm_dM(Rect(0,History.rows,F,2));
	Mat C;
	
	if (TrackerParams->centerHistoryMatrix) {
		Mat B;
		reduce(A, B, 1, CV_REDUCE_SUM, CV_64FC1);
     	C = (TrackerParams->penaltyCoefficient)*(A - (1.0/((double)F))*repeat(B, 1, F));
     }
     else {
     	C = (TrackerParams->penaltyCoefficient)*A;
     }
     
     //Copy elements of C into DP in correct order
     DP = Mat(1, 2*F, CV_64FC1);
     int x = 0;
     int y = 0;
     for (int n = 0; n<2*F; n++) {
     	DP.at<double>(n) = C.at<double>(x, y);
     	x++;
     	if (x == 2) {
     		x = 0;
     		y++;
     	}
     }
}

double compute_P_ExpFact(cv::Mat& xbar, cv::Mat& History, MultiTrackerParameters *TrackerParams) {
	Mat M = build_M(xbar, History);
	if (TrackerParams->centerHistoryMatrix) center_M(M);
	
	//Compute singular values of M
	Mat sigma;
	SVD SVDObject;
	SVDObject.compute(M, sigma, SVD::NO_UV | SVD::MODIFY_A);
	
	double P = 0.0;
	
	int dim = 0;
	if (TrackerParams->centerHistoryMatrix)
		dim = 3;
	else
		dim = 4;
	if (M.cols < dim)
		dim = M.cols;
	
	for (int n=dim; n<sigma.rows; n++)
		P += sigma.at<double>(n);
	return((TrackerParams->penaltyCoefficient)*P);
}

void compute_P_ExpFact(cv::Mat& xbar, cv::Mat& History, double& P, cv::Mat& DP, MultiTrackerParameters *TrackerParams) {
	double sigmaThreshold = 0.05;
	int F = (xbar.rows)/2;
	
	Mat M = build_M(xbar, History);
	if (TrackerParams->centerHistoryMatrix) center_M(M);
	
	//Compute singular value decomposition of M
	Mat sigma, U, Vt;
	SVD SVDObject;
	SVDObject.compute(M, sigma, U, Vt, SVD::MODIFY_A);
	
	int dim = 0;
	if (TrackerParams->centerHistoryMatrix)
		dim = 3;
	else
		dim = 4;
	if (M.cols < dim)
		dim = M.cols;
	
	//Evaluate P
	P = 0.0;
	for (int n=dim; n<sigma.rows; n++)
		P += sigma.at<double>(n);
	P = P*(TrackerParams->penaltyCoefficient);
	
	//Evaluate DP
	for (int n=0; n<dim; n++)
		sigma.at<double>(n) = 0.0;
	for (int n=dim; n<sigma.rows; n++)
		sigma.at<double>(n) = transform_sigma_NucNorm(sigma.at<double>(n), sigmaThreshold);
	Mat Sigma = Mat::diag(sigma);
	
	Mat dSum_dM = U*Sigma*Vt;
	
	Mat A = dSum_dM(Rect(0,History.rows,F,2));
	Mat C;
	
	if (TrackerParams->centerHistoryMatrix) {
		Mat B;
		reduce(A, B, 1, CV_REDUCE_SUM, CV_64FC1);
     	C = (TrackerParams->penaltyCoefficient)*(A - (1.0/((double)F))*repeat(B, 1, F));
     }
     else {
     	C = (TrackerParams->penaltyCoefficient)*A;
     }
     
     //Copy elements of C into DP in correct order
     DP = Mat(1, 2*F, CV_64FC1);
     int x = 0;
     int y = 0;
     for (int n = 0; n<2*F; n++) {
     	DP.at<double>(n) = C.at<double>(x, y);
     	x++;
     	if (x == 2) {
     		x = 0;
     		y++;
     	}
     }
}


//Compute a single fit term - with derivative
void compute_fitTerm(ImageStack& Template, ImageStack& frameStack, int resolutionLevel, Vec2f location, int n, double& residual, double& Dx, double& Dy) {
	       residual     = compute_fitTerm(Template, frameStack, resolutionLevel, location,                   n);
	double c_fit_xPlus  = compute_fitTerm(Template, frameStack, resolutionLevel, location + Vec2f(0.25, 0.0), n);
	double c_fit_xMinus = compute_fitTerm(Template, frameStack, resolutionLevel, location - Vec2f(0.25, 0.0), n);
	double c_fit_yPlus  = compute_fitTerm(Template, frameStack, resolutionLevel, location + Vec2f(0.0, 0.25), n);
	double c_fit_yMinus = compute_fitTerm(Template, frameStack, resolutionLevel, location - Vec2f(0.0, 0.25), n);
	
	Dx = 2.0*(c_fit_xPlus - c_fit_xMinus);
	Dy = 2.0*(c_fit_yPlus - c_fit_yMinus);
}


double compute_fitTerm(ImageStack& Template, ImageStack& frameStack, int resolutionLevel, Vec2f location, int n) {
	//The provided location must be in image coords at the correct resolution level.
	Mat I, imageDelta, temp;
	if ((location[0] >= 0) && (location[0] < frameStack.Images[resolutionLevel].rows) && (location[1] >= 0) && (location[1] < frameStack.Images[resolutionLevel].cols))
		getRectSubPix(frameStack.Images[resolutionLevel], Size(n,n), Point2f(location[1], location[0]), I);
	else
		I = Mat::zeros(n, n, frameStack.Images[resolutionLevel].type());
	return(computeL1Diff_SIMD(Template.Images[resolutionLevel], I)/(n*n));
}


//Compute global energy on a given resolution level, for a given state. No derivative is computed.
double compute_c(Mat& xbar, FeatureSet& features, ImageStack& frameStack, int resolutionLevel, Mat& History, MultiTrackerParameters *TrackerParams) {
	//History is a float matrix. xbar is a float matrix with 1 column.
	//Inputs History and xbar must be represented at the correct resolution level.
	double c = 0.0;
	double P = 0.0;
	
	//Compute contributions of all fit terms
	if (TrackerParams->fitTermEnabled) {
		Vec2f location;
		int f = 0;
		for (list< Feature >::iterator featureIter = features.Features.begin(); featureIter != features.Features.end(); featureIter++) {
			location = Vec2f((double)(xbar.at<float>(2*f)), (double)(xbar.at<float>(2*f+1)));
			c += compute_fitTerm(featureIter->Template, frameStack, resolutionLevel, location, parameters::n);
			f++;
		}
		c = c / ((double) parameters::maxIntensity);
	}
	P = compute_P(xbar, History, TrackerParams); //The penalty coefficient is applied in the compute_P function
	return(c + P);
}

//Compute global energy on a given resolution level, for a given state. The derivative is also computed.
void compute_c(Mat& xbar, FeatureSet& features, ImageStack& frameStack, int resolutionLevel, Mat& History, double& c, Mat& Dc, MultiTrackerParameters *TrackerParams) {
	//History is a float matrix. xbar is a float matrix with 1 column.
	//Inputs History and xbar must be represented at the correct resolution level.
	int F = features.size();
	
	//Declare energy and jacobian component variables
	double c_fit = 0.0;
	double P = 0.0;
	Mat Dc_fit, DP;
	
	//Allocate Dc, if target is not the correct size/type
	if ((Dc.rows != 1) || (Dc.cols != 2*F) || (Dc.type() != CV_64FC1))
		Dc = Mat(1, 2*F, CV_64FC1);
	
	//Compute contributions of all fit terms
	if (TrackerParams->fitTermEnabled) {
		//fprintf(stderr,"\nEvaluating Fit Terms");
		Dc_fit = Mat(1, 2*F, CV_64FC1);
		
		int F = features.size();
		vector<list< Feature >::iterator> featureIters;
		featureIters.resize(F);
		
		//Build vector of feature iterators for parallel execution.
		int f = 0;
		for (list< Feature >::iterator featureIter = features.Features.begin(); featureIter != features.Features.end(); featureIter++) {
			featureIters[f] = featureIter;
			f++;
		}
		
		//Iterate through all features and compute fit term contributions
		for (int f=0; f<F; f++) {
			Vec2f location = Vec2f((double)(xbar.at<float>(2*f)), (double)(xbar.at<float>(2*f+1)));
			double residual;
			compute_fitTerm(featureIters[f]->Template, frameStack, resolutionLevel, location, parameters::n, residual, Dc_fit.at<double>(2*f), Dc_fit.at<double>(2*f+1));
			c_fit += residual;
		}
		
		double alpha = 1.0 / ((double) parameters::maxIntensity);
		c_fit  = alpha * c_fit;
		Dc_fit = alpha * Dc_fit;
	}
	if (TrackerParams->penaltyForm != MultiTrackerParameters::penaltyType::None) {
		//The penalty coefficient is applied in the compute_P function
		compute_P(xbar, History, P, DP, TrackerParams);
	}
	
	//Assemble the total energy and jacobian
	if ((TrackerParams->fitTermEnabled) && (TrackerParams->penaltyForm != MultiTrackerParameters::penaltyType::None)) {
		c = c_fit + P;
		Dc = Dc_fit + DP;
	}
	else if (TrackerParams->fitTermEnabled) {
		c = c_fit;
		Dc = Dc_fit;
	}
	else if (TrackerParams->penaltyForm != MultiTrackerParameters::penaltyType::None) {
		c = P;
		Dc = DP;
	}
	else {
		c = 0.0;
		Dc = Mat::zeros(1, 2*F, CV_64FC1);
	}
}

static Mat build_M(Mat& xbar, Mat& History) {
	//Build and return the matrix M, from the history matrix and xbar.
	//M will not be centered.
	
	int F = (xbar.rows)/2;
	Mat M = Mat(History.rows+2, F, CV_64FC1);
	
	//Copy History matrix into M
	for (int x=0; x<History.rows; x++) {
		for (int y=0; y<F; y++) {
			M.at<double>(x,y) = (double)(History.at<float>(x,y));
		}
	}
	
	//Copy xbar into M
	int x = History.rows;
	int y = 0;
	for (int n=0; n<xbar.rows; n++) {
		M.at<double>(x,y) = (double)(xbar.at<float>(n));
		x++;
		if (x == M.rows) {
			x = History.rows;
			y++;
		}
	}
	
	return(M);
}

static void center_M(cv::Mat& M) {
	//Subtract the mean of all columns from each column of M. This is done in-place.
	
	//Evaluate Centroid
	Mat centroid;
	reduce(M, centroid, 1, CV_REDUCE_AVG, CV_64FC1);

	//Remove Centroid from each column of M
	for (int x=0; x<M.rows; x++) {
		for (int y=0; y<M.cols; y++) {
			M.at<double>(x,y) -= centroid.at<double>(x);
		}
	}
}

static double transform_sigma_EmpDim(double sigma, double epsilon, double delta, double C1, double C2, double threshold) {
	if (sigma > threshold)
		return(C1*pow(sigma, epsilon-1) - C2*pow(sigma, delta-1));
	else
		return((sigma/threshold)*(C1*pow(threshold, epsilon-1) - C2*pow(threshold, delta-1)));
}

static double transform_sigma_NucNorm(double sigma, double threshold) {
	if (sigma > threshold)
		return(1.0);
	else
		return(sigma/threshold);
}
