//This module defines the ImageStack type. An ImageStack is a collection of multiple images,
//understood to correspond to different resolution levels. The different levels of the stack
//may have equal or different sizes. For instance, a standard image pyramid can be built for
//an entire frame of video and saved as an image stack. Alternatively, one can extract n-x-n
//sub-images from the different levels of an image pyramid and save them in an image stack.
//In this case, each level has the same dimensions. In any case, level m in the stack must
//correspond to a standard reference resolution divided down by 2 m times. Thus, at level 0,
//the layer should be a full-resolution image, or a sub-image taken from a full-res image.

#pragma once
#include "opencv2/opencv.hpp"
#include <vector>

class ImageStack {
	public:
		std::vector< cv::Mat > Images;
		
		//Return the number of resolutions levels in the stack.
		int levels(void) const;
		
		//Populate stack by building a partial image pyramid out of a provided image. Save each full level
		//of the pyramid to a level of the stack.
		void BuildFromImage(const cv::Mat& SourceImage, int levels);
		
		//Populate stack by extracting sub-images of size width-by-height from each level of a given stack.
		//If only width is specified, height defaults to equal the provided width.
		void BuildFromStack(const ImageStack& sourceStack, const cv::Vec2f& location, int width, int height=-1);
		
		//Convert image coordinates between resolution levels.
		static void convertImageCoordinates(const cv::Vec2f& inputCoords, cv::Vec2f& outputCoords, int inputLevel, int outputLevel);
};
