#include "Tracker.h"
#include "parameters.h"
#include "EnergyEvaluation.h"
#include "Utilities.h"
#include "prettyPrint.h"

using namespace std;
using namespace cv;
using namespace Eigen;

void Tracker::initialize(int firstFrameNumber, ImageStack firstFrameStack, int FeatureManagementMode, string inputFilePath) {
	frameNumber = firstFrameNumber;
	lastFrameStack = firstFrameStack;
	thisFrameStack = firstFrameStack;
	featureManagementMode = FeatureManagementMode;
	
	if (featureManagementMode == 0)
		reviewFeatureSet();
	else if (featureManagementMode == 1) {
		//Initialize features using initial locations in file
		list<Vec2f> initPositions;
		Utilities::loadInitialPositions(inputFilePath, initPositions);
		
		for (list<Vec2f>::iterator iter = initPositions.begin(); iter != initPositions.end(); iter++) {
			Feature newFeature;
			newFeature.initializeFeature(firstFrameNumber, firstFrameStack, *(iter));
			features.addFeature(newFeature);
		}
		
		//We don't need to do an initial review of features in mode 1 (we never add new features)
	}
	else if (featureManagementMode == 2) {
		reinitializationCounter = 0; //Clear re-initialization counter
		
		//Initialize features and load groundtruth trajectories for each feature
		list<int> birthFrameNumbers;
		list<list<Vec2f>> trajectories;
		if (Utilities::loadTrajectories(inputFilePath, birthFrameNumbers, trajectories) != 0) {
			fprintf(stderr,"Error initializing tracker. Could not load ground-truth file.\n\r");
			return;
		}
		
		int highestBirthFrame = 0;
		for (list<int>::iterator it=birthFrameNumbers.begin(); it != birthFrameNumbers.end(); it++) {
			if (*it > highestBirthFrame)
				highestBirthFrame = *it;
		}
		
		groundTruth_Trajectories.resize(highestBirthFrame + 1);
		list<int>::iterator frameNumber_it = birthFrameNumbers.begin();
		list<list<Vec2f>>::iterator trajectory_it = trajectories.begin();
		while (frameNumber_it != birthFrameNumbers.end()) {
			groundTruth_Trajectories[*frameNumber_it].push_back(*trajectory_it);
			frameNumber_it++;
			trajectory_it++;
		}
		
		//Now review feature set to generate the initial set of features
		reviewFeatureSet();
	}
	else
		fprintf(stderr,"Unrecognized featureManagementMode. Skipping feature initialization.\n\r");
}

//Review the current set of features. Depending on the feature management mode, this function does very different
//things. See the mode definitions for details.
void Tracker::reviewFeatureSet(void) {
	switch(featureManagementMode) {
		case 0:  reviewFeatureSet_mode0(); break;
		case 1:  reviewFeatureSet_mode1(); break;
		case 2:  reviewFeatureSet_mode2(); break;
		default: fprintf(stderr,"Unrecognized featureManagementMode. Skipping review of feature set.\n\r");
	}
}

void Tracker::reviewFeatureSet_mode0(void) {
	int X = thisFrameStack.Images[0].rows;
	int Y = thisFrameStack.Images[0].cols;
	
	Mat cornerScoreMap;
	double harrisK = 0.04; //OpenCV goodFeaturesToTrack Default
	int blockSize = 3;     //OpenCV goodFeaturesToTrack Default

	Mat inputImage;
	if (parameters::M == 0) {
		if (thisFrameStack.Images[0].channels() == 1)
			inputImage = thisFrameStack.Images[0];
		else if (thisFrameStack.Images[0].channels() == 3)
			cvtColor(thisFrameStack.Images[0], inputImage, CV_RGB2GRAY);
		else
			fprintf(stderr,"\nInput image to findNewFeatures must have type CV_8UC1 or CV_8UC3.");
		cornerHarris(inputImage, cornerScoreMap, blockSize, 3, harrisK);
	}
	else {
		if (thisFrameStack.Images[1].channels() == 1)
			inputImage = thisFrameStack.Images[1];
		else if (thisFrameStack.Images[1].channels() == 3)
			cvtColor(thisFrameStack.Images[1], inputImage, CV_RGB2GRAY);
		else
			fprintf(stderr,"\nInput image to findNewFeatures must have type CV_8UC1 or CV_8UC3.");
		Mat lowResScoreMap;
		cornerHarris(inputImage, lowResScoreMap, blockSize, 3, harrisK);
		pyrUp(lowResScoreMap, cornerScoreMap, thisFrameStack.Images[0].size());
	}
	cornerScoreMap = abs(cornerScoreMap);
	
	//Update each features current score
	for (list< Feature >::iterator featureIter = features.Features.begin(); featureIter != features.Features.end(); featureIter++) {
		int x = (int)round(featureIter->CurrentLocation[0]);
		int y = (int)round(featureIter->CurrentLocation[1]);
		if (x < 0) x=0;
		if (x > X-1) x=X-1;
		if (y < 0) y=0;
		if (y > Y-1) y=Y-1;
		
		//The constant alpha dictates how rapidly a features score adjusts to new imagery.
		const float alpha = 0.05;
		featureIter->currentScore = alpha*cornerScoreMap.at<float>(x,y) + (1.0 - alpha)*featureIter->currentScore;
	}
	
	//Review current features
	list< Feature >::iterator featureIter = features.Features.begin();
	while (featureIter != features.Features.end()) {
		//Check to see if the feature is too close to an image edge or if the features score is insufficient for tracking
		double x = featureIter->CurrentLocation[0];
		double y = featureIter->CurrentLocation[1];
		float featureScore = featureIter->currentScore;
		if ((x < parameters::oldFeatureEdgeRadius) || (x >= (double)X - parameters::oldFeatureEdgeRadius) ||
		    (y < parameters::oldFeatureEdgeRadius) || (y >= (double)Y - parameters::oldFeatureEdgeRadius) ||
		    (featureScore < parameters::minFeatureScore)) {
			//Feature needs to be removed
			deadFeatures.addFeature(*featureIter);
			features.removeFeature(featureIter);
		}
		else
			featureIter++;
	}
	
	if (parameters::extensiveFeatureReviewPeriod > 0)
		extensiveReviewOfFeatures(cornerScoreMap);
	
	//Add new features, if necessary.
	int numberOfNewFeatures = parameters::numberOfFeatures - features.size();
	if (numberOfNewFeatures > 0) {
		list<Vec2f> newFeatureLocations;
		findNewFeatures(numberOfNewFeatures, cornerScoreMap, newFeatureLocations);
		for (list< Vec2f >::iterator newFeatureIter = newFeatureLocations.begin(); newFeatureIter != newFeatureLocations.end(); newFeatureIter++) {
			Feature newFeature;
			newFeature.initializeFeature(frameNumber, thisFrameStack, *(newFeatureIter));
			features.addFeature(newFeature);
		}
	}
}

void Tracker::extensiveReviewOfFeatures(Mat& cornerScoreMap) {
	//Exhaustive reviews are being conducted periodically - we make sure we are tracking the best features in the scene.
	if ((parameters::extensiveFeatureReviewPeriod > 0) && (frameNumber % parameters::extensiveFeatureReviewPeriod == 0)) {
		//It is time to perform an exhaustive review of our feature set
		list<Vec2f> TopFeatureLocations;
	
		int X = thisFrameStack.Images[0].rows;
		int Y = thisFrameStack.Images[0].cols;
		int cellSize = (X + Y)/20;

		int XNumCells, YNumCells;
		if (X % cellSize == 0)
			XNumCells = X/cellSize;
		else
			XNumCells = X/cellSize + 1;
		if (Y % cellSize == 0)
			YNumCells = Y/cellSize;
		else
			YNumCells = Y/cellSize + 1;
		Mat gridScoreMap = Mat::zeros(XNumCells, YNumCells, CV_32FC1);
		Mat grid_x       = Mat(XNumCells, YNumCells, CV_32SC1);
		Mat grid_y       = Mat(XNumCells, YNumCells, CV_32SC1);

		//For each cell in the coarse grid, save the best scoring feature location and score in that cell.
		int gridx, gridy;
		for (int x=(parameters::newFeatureEdgeRadius); x<(X-parameters::newFeatureEdgeRadius); x++) {
			for (int y=(parameters::newFeatureEdgeRadius); y<(Y-parameters::newFeatureEdgeRadius); y++) {
				gridx = x/cellSize;
				gridy = y/cellSize;
				if (cornerScoreMap.at<float>(x,y) > gridScoreMap.at<float>(gridx, gridy)) {
					gridScoreMap.at<float>(gridx, gridy) = cornerScoreMap.at<float>(x,y);
					grid_x.at<int32_t>(gridx, gridy) = x;
					grid_y.at<int32_t>(gridx, gridy) = y;
				}
			}
		}

		//Now reshape the matrices into linear buffers and sort the scores
		gridScoreMap = gridScoreMap.reshape(0, 1);
		grid_x       = grid_x.reshape(0, 1);
		grid_y       = grid_y.reshape(0, 1);
		Mat Indices;
		sortIdx(gridScoreMap, Indices, CV_SORT_EVERY_ROW | CV_SORT_DESCENDING);

		//Populate new feature list with best new candidates
		TopFeatureLocations.clear();
		for (int n=0; n<Utilities::min_int(gridScoreMap.cols, parameters::numberOfTopFeaturesThatMustBeTracked); n++) {
			int index = Indices.at<int32_t>(n);
			TopFeatureLocations.push_back(Vec2f(grid_x.at<int32_t>(index),grid_y.at<int32_t>(index)));
		}
	
		for (list< Vec2f >::iterator LocIter = TopFeatureLocations.begin(); LocIter != TopFeatureLocations.end(); LocIter++) {
			float distance = features.getDistanceFromTrackset(*LocIter);
			if (distance > parameters::identificationTolerance) {
				//We are not currently tracking this feature
				if (features.size() >= parameters::numberOfFeatures) {
					//Before adding new feature, throw out existing worst feature
				
					//First score all features and identify the worst current feature
					float featureScore, worstScore = 0.0;
					list< Feature >::iterator worstFeatureIter;
					
					for (list< Feature >::iterator featureIter = features.Features.begin(); featureIter != features.Features.end(); featureIter++) {
						int x = (int)round(featureIter->CurrentLocation[0]);
						int y = (int)round(featureIter->CurrentLocation[1]);
						if (x < 0) x=0;
						if (x > X-1) x=X-1;
						if (y < 0) y=0;
						if (y > Y-1) y=Y-1;
						
						if (featureIter == features.Features.begin()) {
							worstScore = cornerScoreMap.at<float>(x,y);
							worstFeatureIter = featureIter;
						}
						else {
							featureScore = cornerScoreMap.at<float>(x,y);
							if (featureScore < worstScore) {
								worstScore = featureScore;
								worstFeatureIter = featureIter;
							}
						}
					}
				
					//Remove the feature with the lowest score
					deadFeatures.addFeature(*worstFeatureIter);
					features.removeFeature(worstFeatureIter);
				}
			
				//Add new feature to trackset
				Feature newFeature;
				newFeature.initializeFeature(frameNumber, thisFrameStack, *(LocIter));
				features.addFeature(newFeature);
			}
		}
	}
}

void Tracker::reviewFeatureSet_mode1(void) {
	//In this mode, we simply throw out features that get too close to leaving the field of view
	int X = thisFrameStack.Images[0].rows;
	int Y = thisFrameStack.Images[0].cols;
	list< Feature >::iterator featureIter = features.Features.begin();
	while (featureIter != features.Features.end()) {
		//Check to see if the feature is too close to an image edge
		double x = featureIter->CurrentLocation[0];
		double y = featureIter->CurrentLocation[1];
		if ((x < parameters::oldFeatureEdgeRadius) || (x >= (double)X - parameters::oldFeatureEdgeRadius) ||
		    (y < parameters::oldFeatureEdgeRadius) || (y >= (double)Y - parameters::oldFeatureEdgeRadius)) {
			//Feature needs to be removed
			deadFeatures.addFeature(*featureIter);
			features.removeFeature(featureIter);
		}
		else
			featureIter++;
	}
}

void Tracker::reviewFeatureSet_mode2(void) {
	//In this mode, we see if there are eny new features that need to be initialized and we initialize them. Then, we
	//go through our feature set and re-initialize those features that have drifted too far from ground truth and we
	//kill those features that are supposed to have terminated (according to ground truth).
	
	//First, add features that are supposed to be born this frame
	if (frameNumber < ((int) groundTruth_Trajectories.size())) {
		list<list<Vec2f>> newTrajectories = groundTruth_Trajectories[frameNumber];
		for (list<list<Vec2f>>::iterator it = newTrajectories.begin(); it != newTrajectories.end(); it++) {
			list<Vec2f> newTrajectory = *it;
			
			Feature newFeature;
			newFeature.initializeFeature(frameNumber, thisFrameStack, newTrajectory.front());
			newFeature.loadGroundTruth(newTrajectory, frameNumber);
			features.addFeature(newFeature);
		}
	}
	
	//Next, kill any features than are not supposed to exist anymore
	list< Feature >::iterator featureIter = features.Features.begin();
	while (featureIter != features.Features.end()) {
		if (!featureIter->groundTruth_Alive()) {
			deadFeatures.addFeature(*featureIter);
			features.removeFeature(featureIter);
		}
		else
			featureIter++;
	}
	
	//Re-initialize featureas that have wandered too far from ground truth
	for (list< Feature >::iterator featureIter = features.Features.begin(); featureIter != features.Features.end(); featureIter++) {
		if (featureIter->getDeviationFromGroundTruth() > featureReinitTolerance) {
			//Re-initialize feature - including updating the template
			featureIter->CurrentLocation = featureIter->groundTruth_Location(featureIter->CurrentFrameNumber);
			
			featureIter->Template.BuildFromStack(thisFrameStack, featureIter->CurrentLocation, parameters::n);
			featureIter->RecentTemplateCandidates.clear();
			featureIter->RecentTemplateCandidates.push_front(featureIter->Template);
			
			//Record this re-initialization event
			reinitializationCounter++;
		}
	}
}


void Tracker::trackToNextFrame_KLT(int nextFrameNumber, ImageStack nextFrameStack) {
	//First, update frame stacks and frame number
	updateFrameNumberAndStacks(nextFrameNumber, nextFrameStack);
	
	if (features.size() > 0) {
		Mat prev_grey, current_grey;
		if (thisFrameStack.Images[0].type() == CV_8UC3) {
			cvtColor(lastFrameStack.Images[0], prev_grey,    CV_RGB2GRAY);
			cvtColor(thisFrameStack.Images[0], current_grey, CV_RGB2GRAY);
		}
		else {
			prev_grey    = lastFrameStack.Images[0];
			current_grey = thisFrameStack.Images[0];
		}
		vector<Point2f> inputFeatureCoords;
		inputFeatureCoords.resize(features.size());
		int f = 0;
		for (list< Feature >::iterator featureIter = features.Features.begin(); featureIter != features.Features.end(); featureIter++) {
			inputFeatureCoords[f].y = featureIter->CurrentLocation[0];
			inputFeatureCoords[f].x = featureIter->CurrentLocation[1];
			f++;
		}
		vector<Point2f> outputFeatureCoords;
		vector<uchar> status;
		vector<float> error;
		calcOpticalFlowPyrLK(prev_grey, current_grey, inputFeatureCoords, outputFeatureCoords, status, error, Size(parameters::n, parameters::n), parameters::M);
		f = 0;
		Vec2f location;
		for (list< Feature >::iterator featureIter = features.Features.begin(); featureIter != features.Features.end(); featureIter++) {
			location = Vec2f(outputFeatureCoords[f].y, outputFeatureCoords[f].x);
			featureIter->Update(frameNumber, thisFrameStack, location);
			f++;
		}
	}
	reviewFeatureSet();
}


void Tracker::trackToNextFrame_1stOrderDescent(int nextFrameNumber, ImageStack nextFrameStack) {
	//First, update frame stacks and frame number
	updateFrameNumberAndStacks(nextFrameNumber, nextFrameStack);
	
	//Track each feature and update as you go
	Vec2f location;
	int f = 0;
	for (list< Feature >::iterator featureIter = features.Features.begin(); featureIter != features.Features.end(); featureIter++) {
		location = trackSingleFeature_1stOrderDescent(*featureIter, nextFrameStack, featureIter->CurrentLocation);
		featureIter->Update(frameNumber, thisFrameStack, location);
		f++;
	}
	reviewFeatureSet();
}


void Tracker::trackToNextFrame_KLT_GlobalMotionInit(int nextFrameNumber, ImageStack nextFrameStack) {
	//First, update frame stacks and frame number
	updateFrameNumberAndStacks(nextFrameNumber, nextFrameStack);
	
	if (features.size() > 0) {
		//Estimate Global Motion
		float globalMotion_x = 0.0;
		float globalMotion_y = 0.0;
		estimateGlobalMotion(globalMotion_x, globalMotion_y);
		Vec2f globalMotion(globalMotion_x, globalMotion_y);
	
		Mat prev_grey, current_grey;
		if (thisFrameStack.Images[0].type() == CV_8UC3) {
			cvtColor(lastFrameStack.Images[0], prev_grey,    CV_RGB2GRAY);
			cvtColor(thisFrameStack.Images[0], current_grey, CV_RGB2GRAY);
		}
		else {
			prev_grey    = lastFrameStack.Images[0];
			current_grey = thisFrameStack.Images[0];
		}
		vector<Point2f> inputFeatureCoords;
		inputFeatureCoords.resize(features.size());
		int f = 0;
		for (list< Feature >::iterator featureIter = features.Features.begin(); featureIter != features.Features.end(); featureIter++) {
			inputFeatureCoords[f].y = featureIter->CurrentLocation[0] + globalMotion_x;
			inputFeatureCoords[f].x = featureIter->CurrentLocation[1] + globalMotion_y;
			f++;
		}
		vector<Point2f> outputFeatureCoords;
		vector<uchar> status;
		vector<float> error;
		calcOpticalFlowPyrLK(prev_grey, current_grey, inputFeatureCoords, outputFeatureCoords, status, error, Size(parameters::n, parameters::n), parameters::M);
		f = 0;
		Vec2f location;
		for (list< Feature >::iterator featureIter = features.Features.begin(); featureIter != features.Features.end(); featureIter++) {
			location = Vec2f(outputFeatureCoords[f].y, outputFeatureCoords[f].x);
			featureIter->Update(frameNumber, thisFrameStack, location);
			f++;
		}
	}
	reviewFeatureSet();
}

void Tracker::trackToNextFrame_1stOrderDescent_GlobalMotionInit(int nextFrameNumber, ImageStack nextFrameStack) {
	//First, update frame stacks and frame number
	updateFrameNumberAndStacks(nextFrameNumber, nextFrameStack);
	
	//Estimate Global Motion
	float globalMotion_x = 0.0;
	float globalMotion_y = 0.0;
	estimateGlobalMotion(globalMotion_x, globalMotion_y);
	Vec2f globalMotion(globalMotion_x, globalMotion_y);
	
	//Track each feature and update as you go
	Vec2f location;
	int f = 0;
	for (list< Feature >::iterator featureIter = features.Features.begin(); featureIter != features.Features.end(); featureIter++) {
		location = trackSingleFeature_1stOrderDescent(*featureIter, nextFrameStack, featureIter->CurrentLocation + globalMotion);
		featureIter->Update(frameNumber, thisFrameStack, location);
		f++;
	}
	reviewFeatureSet();
}


void Tracker::trackToNextFrame_multiTrack_GlobalMotionInit(int nextFrameNumber, ImageStack nextFrameStack) {
	//First, update frame stacks and frame number
	updateFrameNumberAndStacks(nextFrameNumber, nextFrameStack);
	
	if (features.size() > 0) {
		//Estimate Global Motion
		float globalMotion_x = 0.0;
		float globalMotion_y = 0.0;
		estimateGlobalMotion(globalMotion_x, globalMotion_y);

		//Generate xbar and history matrix
		Mat xbar, History;
		features.build_xbar(xbar);
		features.build_HistoryMatrix(History, multiTrackerParametersPtr->L);

		//Add estimated global motion to xbar
		for (int f=0; f<features.size(); f++) {
			xbar.at<float>(2*f)   += globalMotion_x;
			xbar.at<float>(2*f+1) += globalMotion_y;
		}

		//Track on coarsest resolution and refine on higher-res images
		for (int m=(parameters::M-1); m>=0; m--)
			ConstrainedMultiTracker_TrackOnSingleLevel(xbar, thisFrameStack, m, History);

		//xbar now holds best tracker state at resolution level 0. Update features with new locations
		Vec2f location;
		int f = 0;
		for (list< Feature >::iterator featureIter = features.Features.begin(); featureIter != features.Features.end(); featureIter++) {
			location = Vec2f(xbar.at<float>(2*f) , xbar.at<float>(2*f+1));
			featureIter->Update(frameNumber, thisFrameStack, location);
			f++;
		}
	}
	reviewFeatureSet();
}


void Tracker::updateFrameNumberAndStacks(int nextFrameNumber, ImageStack &nextFrameStack) {
	//Update the frame number and frame stacks. This is the first step of tracking to a new frame
	frameNumber = nextFrameNumber;
	lastFrameStack = thisFrameStack;
	thisFrameStack = nextFrameStack;
}


Vec2f Tracker::trackSingleFeature_1stOrderDescent(Feature& f, ImageStack& nextFrameStack, Vec2f initialLocationEstimate) {
	Vec2f location = initialLocationEstimate;
	Vec2f state, step, testLocation;
	double c = 0.0;
	double Dx = 0.0;
	double Dy = 0.0;
	double testa, testb, testc;
	double stepNorm = 0.0;
	double lastStepNorm = 0.0;
	int minSteps, maxSteps;
	
	for (int m=(parameters::M-1); m>=0; m--) {
		ImageStack::convertImageCoordinates(location, state, 0, m);
		
		if (m == parameters::M-1) {
			minSteps = parameters::minStepsTopLevel;
			maxSteps = parameters::maxStepsTopLevel;
		}
		else {
			minSteps = parameters::minStepsLowerLevels;
			maxSteps = parameters::maxStepsLowerLevels;
		}
		
		for (int stepNumber = 0; stepNumber < maxSteps; stepNumber++) {
			compute_fitTerm(f.Template, nextFrameStack, m, state, parameters::n, c, Dx, Dy);
			step = -1.0*Vec2f((float)Dx ,(float)Dy);
			stepNorm = norm(step);
			if ((stepNumber > 0) && (stepNumber >= minSteps) && (stepNorm > 0.9999*lastStepNorm))
				break;
			lastStepNorm = stepNorm;
			if (stepNorm < 0.00001)
				break;
			step = step/stepNorm;
			
			double scale = 2.0;
			
			int maxNumberOfTests = 100;
			int maxScaleRefinements = 10;
		
			int scaleRefinements = 0;
		
			double testaDist = 0.0;
			testa = c;
		
			double testbDist = 1.0*scale;
			testLocation = state + testbDist*step;
			testb = compute_fitTerm(f.Template, nextFrameStack, m, testLocation, parameters::n);
		
			double testcDist = 2.0*scale;
			testLocation = state + testcDist*step;
			testc = compute_fitTerm(f.Template, nextFrameStack, m, testLocation, parameters::n);
		
			for (int testNumber=0; testNumber<maxNumberOfTests; testNumber++) {
				if ((testa > testb) && (testb > testc)) {
					//Keep moving out on this scale
					testaDist = testbDist;
					testa = testb;
					testbDist = testcDist;
					testb = testc;
					testcDist += scale;
					testLocation = state + testcDist*step;
					testc = compute_fitTerm(f.Template, nextFrameStack, m, testLocation, parameters::n);
				}
				else {
					//Shorten scale
					testcDist = testbDist;
					testc = testb;
					scale = (0.5)*scale;
					scaleRefinements ++;
					testbDist = testaDist + scale;
					testLocation = state + testbDist*step;
					testb = compute_fitTerm(f.Template, nextFrameStack, m, testLocation, parameters::n);
				}
			
				//Test break conditions
				if (scaleRefinements >= maxScaleRefinements)
					break;
			}
		
			//Update state
			state = state + testbDist*step;
		}
		
		ImageStack::convertImageCoordinates(state, location, m, 0);
	}
	
	return(location);
}


void Tracker::ConstrainedMultiTracker_TrackOnSingleLevel(Mat& xbar, ImageStack& frameStack, int resolutionLevel, Mat& History) {
	//The inputs xbar and History must be in image coordinates at resolution level 0. xbar will be updated on completion of this function.
	//The updated xbar will also be in image coordinates at resolution level 0.
	
	//Convert xbar and History to the correct resolution level
	double temp1 = pow(2.0, (double)resolutionLevel);
	double temp2 = 1.0/temp1;
	xbar = temp2*xbar;
	Mat History_thisLevel = temp2*History;
	
	int F = features.size();
	
	//Iteratively Descend Energy Surface
	double lastDcNorm = 100.0;
	double DcNorm = 0.0;
	double c, scale;
	Mat Dc, step, step_1, step_2, stepDouble, nextxbar, testLocation;
	Mat lastStepDir = Mat::zeros(2*F,1,CV_64FC1);
	int maxStepsThisLevel, minStepsThisLevel;
	if (resolutionLevel == parameters::M-1) {
		maxStepsThisLevel = parameters::maxStepsTopLevel;
		minStepsThisLevel = parameters::minStepsTopLevel;
	}
	else {
		maxStepsThisLevel = parameters::maxStepsLowerLevels;
		minStepsThisLevel = parameters::minStepsLowerLevels;
	}
	for (int stepNumber = 0; stepNumber < maxStepsThisLevel; stepNumber++) {
		nextxbar = xbar;
		
		compute_c(nextxbar, features, frameStack, resolutionLevel, History_thisLevel, c, Dc, multiTrackerParametersPtr);
		//Semi-standardize derivatives
		step_1 = -1.0*Dc.t();
		step_1.copyTo(step_2);
		for (int f=0; f<F; f++) {
			double subNorm = sqrt(pow(step_2.at<double>(2*f), 2.0) + pow(step_2.at<double>(2*f+1), 2.0));
			if (subNorm > 0.000001) {
				step_2.at<double>(2*f) /= sqrt(subNorm);
				step_2.at<double>(2*f+1) /= sqrt(subNorm);
			}
		}
		stepDouble = 0.5*step_1 + 0.5*step_2;
		double maxVal = abs(stepDouble.at<double>(0));
		for (int i=1; i<2*F; i++)
			if (abs(stepDouble.at<double>(i)) > maxVal)
				maxVal = abs(stepDouble.at<double>(i));
		stepDouble.convertTo(step, CV_32F, 1.0/maxVal);
		
		//Test early exit conditions
		DcNorm = norm(Dc);
		if (stepNumber >= minStepsThisLevel) {
			if (DcNorm > 0.99*lastDcNorm) {
				//fprintf(stderr,"\nTerminating because |DC| is not dropping anymore. Ran %d steps.", stepNumber);
				break;
			}
			else if (DcNorm < parameters::DerivativeTolerance) {
				//fprintf(stderr,"\nTerminating because |DC| < DerivativeTolerance. Ran %d steps.", stepNumber);
				break;
			}
		}
		lastDcNorm = DcNorm;
		
		//if (resolutionLevel <= parameters::M-1) {
		if (false) {
			//scale = 20.0;
			scale = 2.0;
			xbar = nextxbar + scale*step;
		}
		else {
			//On finest scale, perform precision line search
			scale = 10.0;
			
			int maxNumberOfTests = 100;
			int maxScaleRefinements = 10;
		
			int scaleRefinements = 0;
		
			double testaDist = 0.0;
			double testa = c;
		
			double testbDist = 1.0*scale;
			testLocation = nextxbar + testbDist*step;
			double testb = compute_c(testLocation, features, frameStack, resolutionLevel, History_thisLevel, multiTrackerParametersPtr);
		
			double testcDist = 2.0*scale;
			testLocation = nextxbar + testcDist*step;
			double testc = compute_c(testLocation, features, frameStack, resolutionLevel, History_thisLevel, multiTrackerParametersPtr);
		
			for (int testNumber=0; testNumber<maxNumberOfTests; testNumber++) {
				if ((testa > testb) && (testb > testc)) {
					//Keep moving out on this scale
					testaDist = testbDist;
					testa = testb;
					testbDist = testcDist;
					testb = testc;
					testcDist += scale;
					testLocation = nextxbar + testcDist*step;
					testc = compute_c(testLocation, features, frameStack, resolutionLevel, History_thisLevel, multiTrackerParametersPtr);
				}
				else {
					//Shorten scale
					testcDist = testbDist;
					testc = testb;
					scale = (0.5)*scale;
					scaleRefinements ++;
					testbDist = testaDist + scale;
					testLocation = nextxbar + testbDist*step;
					testb = compute_c(testLocation, features, frameStack, resolutionLevel, History_thisLevel, multiTrackerParametersPtr);
				}
			
				//Test break conditions
				if (scaleRefinements >= maxScaleRefinements)
					break;
			}
		
			//Update state
			xbar = nextxbar + testbDist*step;
		}
	}
	
	//Convert xbar back to resolution level 0
	xbar = temp1*xbar;
}

void Tracker::estimateGlobalMotion(float& delta_x, float& delta_y) {
	//Track on coarsest resolution and refine on 1 higher-resolution version
	int finestLevel = parameters::M-2;
	if (finestLevel < 0)
		finestLevel = 0;
	
	Vec2f initialLocation, initialLocation_levelm, location;
	initialLocation[0] = ((float)(thisFrameStack.Images[0].rows - 1))/2.0;
	initialLocation[1] = ((float)(thisFrameStack.Images[0].cols - 1))/2.0;
	location[0] = initialLocation[0];
	location[1] = initialLocation[1];
	
	for (int m=(parameters::M-1); m >= finestLevel; m--) {
		float temp1 = (float)pow(2.0, (double)m);
		float temp2 = 1.0/temp1;
		location = temp2*location;
		initialLocation_levelm = temp2*initialLocation;
		
		double residual, Dx, Dy;
		Vec2f step;
		int n = min2i(thisFrameStack.Images[m].cols-10, thisFrameStack.Images[m].rows-10);
		ImageStack Template;
		Template.Images.resize(parameters::M);
		getRectSubPix(lastFrameStack.Images[m], Size(n,n), Point2f(initialLocation_levelm[1], initialLocation_levelm[0]), Template.Images[m]);
		for (int stepNumber = 0; stepNumber < 10; stepNumber++) {
			//Gradient Decent step
			compute_fitTerm(Template, thisFrameStack, m, location, n, residual, Dx, Dy);
			location = location - 0.1*Vec2f(Dx, Dy);
		}
		
		location = temp1*location;
	}
	
	delta_x = location[0] - initialLocation[0];
	delta_y = location[1] - initialLocation[1];
}

int Tracker::min2i(int a, int b) {
	if (a <= b)
		return(a);
	else
		return(b);
}

void Tracker::findNewFeatures(int numberOfNewFeatures, Mat& cornerScoreMap, list<Vec2f>& newFeatureLocations) {
	int X = thisFrameStack.Images[0].rows;
	int Y = thisFrameStack.Images[0].cols;
	int cellSize = (X + Y)/20;
	
	int XNumCells, YNumCells;
	if (X % cellSize == 0)
		XNumCells = X/cellSize;
	else
		XNumCells = X/cellSize + 1;
	if (Y % cellSize == 0)
		YNumCells = Y/cellSize;
	else
		YNumCells = Y/cellSize + 1;
	Mat gridScoreMap = Mat::zeros(XNumCells, YNumCells, CV_32FC1);
	Mat grid_x       = Mat(XNumCells, YNumCells, CV_32SC1);
	Mat grid_y       = Mat(XNumCells, YNumCells, CV_32SC1);
	Mat grid_Hit     = Mat::zeros(XNumCells, YNumCells, CV_8UC1);
	
	//For each cell in the coarse grid, save the best scoring feature location and score in that cell.
	int gridx, gridy;
	for (int x=(parameters::newFeatureEdgeRadius); x<(X-parameters::newFeatureEdgeRadius); x++) {
		for (int y=(parameters::newFeatureEdgeRadius); y<(Y-parameters::newFeatureEdgeRadius); y++) {
			gridx = x/cellSize;
			gridy = y/cellSize;
			if (cornerScoreMap.at<float>(x,y) > gridScoreMap.at<float>(gridx, gridy)) {
				gridScoreMap.at<float>(gridx, gridy) = cornerScoreMap.at<float>(x,y);
				grid_x.at<int32_t>(gridx, gridy) = x;
				grid_y.at<int32_t>(gridx, gridy) = y;
				grid_Hit.at<uchar>(gridx, gridy) = 1U;
			}
		}
	}
	
	//Invalidate cells that contain existing features.
	for (list< Feature >::iterator featureIter = features.Features.begin(); featureIter != features.Features.end(); featureIter++) {
		//Mark the cell holding the current feature as containing no new features
		gridx = featureIter->CurrentLocation[0]/cellSize;
		gridy = featureIter->CurrentLocation[1]/cellSize;
		
		if (gridx < 0) gridx = 0;
		if (gridx > XNumCells-1) gridx = XNumCells-1;
		if (gridy < 0) gridy = 0;
		if (gridy > YNumCells-1) gridy = YNumCells-1;
		
		grid_Hit.at<uchar>(gridx, gridy) = 0U;
	}
	
	//Now reshape the matrices into linear buffers and sort the scores
	gridScoreMap = gridScoreMap.reshape(0, 1);
	grid_x       = grid_x.reshape(0, 1);
	grid_y       = grid_y.reshape(0, 1);
	grid_Hit     = grid_Hit.reshape(0, 1);
	Mat Indices;
	sortIdx(gridScoreMap, Indices, CV_SORT_EVERY_ROW | CV_SORT_DESCENDING);
	
	//Populate new feature list with best new candidates
	newFeatureLocations.clear();
	int newFeaturesAdded = 0;
	for (int n=0; n<gridScoreMap.cols; n++) {
		int index = Indices.at<int32_t>(n);
		if (grid_Hit.at<uchar>(index) == 1U) {
			//There is a valid new feature here
			newFeatureLocations.push_back(Vec2f(grid_x.at<int32_t>(index),grid_y.at<int32_t>(index)));
			newFeaturesAdded++;
			if (newFeaturesAdded >= numberOfNewFeatures)
				break;
		}
	}
}

