//External Includes
#include "opencv2/opencv.hpp"
#include <stdio.h>
#include <unistd.h>
#include <sys/time.h>

//Internal Includes
#include "Tracker.h"
#include "ImageStack.h"
#include "parameters.h"
#include "Benchmark.h"
#include "Utilities.h"

//#define _GNU_SOURCE 1
#include <fenv.h>

//Set default option values
#define DEFAULT_TRACKERID 0
#define DEFAULT_VIDEOCAM_NUM 0

using namespace cv;
using namespace std;
using namespace Eigen;
using namespace Utilities;

//Run this program with the -h command-line option for usage options.

//Forward Declarations
void getFrameWithTimestamp(VideoCapture &cap, Mat &frame, struct timeval &RXTime);
void displayHelp(void);

//Constants
const double pi = 3.141592653589793;

int main(int argc, char** argv) {
     int key, retval;
     Mat temp, frame, frameBW, frameWithFeatures;
     
     int cameraOpenCVIndex = 0;
     int trackerID;
     
     //Parse command-line options
	int tflag = 0;
	int vflag = 0;
	int hflag = 0;
	char *tval = NULL;
	char *vval = NULL;

	int c;
	opterr = 0; //Do not print an error when unrecognized arguments are encountered

	while ((c = getopt (argc, argv, "t:v:h")) != -1)
		switch (c) {
			case 't':
				tflag = 1;
				tval = optarg;
				break;
			case 'v':
				vflag = 1;
				vval = optarg;
				break;
			case 'h':
				hflag = 1;
				break;
			case '?':
				fprintf (stderr, "Unrecognized Option: -%c found. Aborting.\n\r", optopt);
				return(-1);
			default:
				fprintf(stderr, "Unexpected value returned by getopt. Aborting.\n\r");
				return(-1);
	}
	
	if (tflag == 1)
		trackerID = atoi(tval);
	else
		trackerID = DEFAULT_TRACKERID;
	
	if (vflag == 1)
		cameraOpenCVIndex = atoi(vval);
	else
		cameraOpenCVIndex = DEFAULT_VIDEOCAM_NUM;
	
	//If help flag was passed, print help text and exit imediately
	if (hflag == 1) {
		displayHelp();
		return(0);
	}
     
     //TODO:Disable for non-debug mode
     feenableexcept(FE_INVALID | FE_OVERFLOW);
     
     //Initialize Tracker Output Window
     namedWindow("Live Display", CV_WINDOW_AUTOSIZE|CV_GUI_NORMAL);
     
     //Open Video Camera
	VideoCapture cap(cameraOpenCVIndex); //Open the camera
	if(!cap.isOpened())                  //Check if we succeeded
		return -1;
	
	//Disable autofocus for camera (if desired) - Linux and V4L only
	key = waitForButtonPress(cap, "Live Display", 5, string("Press d to disable autofocus."), string("Press any other key to skip."));
	if (key == 100) {
		char command[64];
		fprintf(stderr,"Deactivating Auto-focus... ");
		sprintf(command, "v4l2-ctl --device=%d --set-ctrl=focus_auto=0", cameraOpenCVIndex);
		retval = system(command);
		if (retval == 0) fprintf(stderr,"Success\n\r");
		else fprintf(stderr,"Failed\n\r");
	}
	else {
		fprintf(stderr,"Skipping Auto-focus deactivation.\n\r");
	}
	
	fprintf(stderr,"Rendering live display.\n\rPress spacebar to exit.\n\r");
	struct timeval newTime;
	Mat newFrame;
	getFrameWithTimestamp(cap, newFrame, newTime);
	ImageStack frameStack;
     Tracker myTracker;
     
     //Build Multi-Feature tracker options struct and set up appropriately
     MultiTrackerParameters myMultiFeatureTrackerParams;
     myMultiFeatureTrackerParams.L = 5;
     switch(trackerID) {
     	case 5:  myMultiFeatureTrackerParams.penaltyForm = MultiTrackerParameters::penaltyType::EmpiricalDimension;
     	         myMultiFeatureTrackerParams.penaltyCoefficient = 0.1;
     	         myMultiFeatureTrackerParams.centerHistoryMatrix = false;
     	         break;
     	case 6:  myMultiFeatureTrackerParams.penaltyForm = MultiTrackerParameters::penaltyType::EmpiricalDimension;
     	         myMultiFeatureTrackerParams.penaltyCoefficient = 0.15;
     	         myMultiFeatureTrackerParams.centerHistoryMatrix = true;
     	         break;
     	case 7:  myMultiFeatureTrackerParams.penaltyForm = MultiTrackerParameters::penaltyType::NuclearNorm;
     	         myMultiFeatureTrackerParams.penaltyCoefficient = 0.001;
     	         myMultiFeatureTrackerParams.centerHistoryMatrix = false;
     	         break;
     	case 8:  myMultiFeatureTrackerParams.penaltyForm = MultiTrackerParameters::penaltyType::NuclearNorm;
     	         myMultiFeatureTrackerParams.penaltyCoefficient = 0.0005;
     	         myMultiFeatureTrackerParams.centerHistoryMatrix = true;
     	         break;
     	case 9:  myMultiFeatureTrackerParams.penaltyForm = MultiTrackerParameters::penaltyType::ExplicitFactorization;
     	         myMultiFeatureTrackerParams.penaltyCoefficient = 0.0015;
     	         myMultiFeatureTrackerParams.centerHistoryMatrix = false;
     	         break;
     	case 10: myMultiFeatureTrackerParams.penaltyForm = MultiTrackerParameters::penaltyType::ExplicitFactorization;
     	         myMultiFeatureTrackerParams.penaltyCoefficient = 0.002;
     	         myMultiFeatureTrackerParams.centerHistoryMatrix = true;
     	         break;
     }
	
	//Initialize Tracker
	frameStack.BuildFromImage(newFrame, parameters::M);
	myTracker.initialize(0, frameStack, 0);
	myTracker.multiTrackerParametersPtr = &myMultiFeatureTrackerParams;
	
     fprintf(stderr,"\nTracker initialized - Tracking input stream.\n");
	
	for (int frameNumber=1; ; frameNumber++) {
		newFrame = Mat();
		getFrameWithTimestamp(cap, newFrame, newTime);
		tic(); //Start timer for detirmining processing frame rate.
		
		//Build frame stack for new frame
		frameStack.BuildFromImage(newFrame, parameters::M);
     	
     	//Track features to new frame
     	switch (trackerID) {
     		case 0:  myTracker.trackToNextFrame_KLT                             (frameNumber, frameStack);    break;
     		case 1:  fprintf(stderr,"Tracker not supported in this release\n");                               return(-1);
     		case 2:  myTracker.trackToNextFrame_1stOrderDescent_GlobalMotionInit(frameNumber, frameStack);    break;
     		case 3:  fprintf(stderr,"Tracker not supported in this release\n");                               return(-1);
     		case 4:  fprintf(stderr,"Tracker not supported in this release\n");                               return(-1);
     		case 5:  myTracker.trackToNextFrame_multiTrack_GlobalMotionInit     (frameNumber, frameStack);    break;
			case 6:  myTracker.trackToNextFrame_multiTrack_GlobalMotionInit     (frameNumber, frameStack);    break;
			case 7:  myTracker.trackToNextFrame_multiTrack_GlobalMotionInit     (frameNumber, frameStack);    break;
			case 8:  myTracker.trackToNextFrame_multiTrack_GlobalMotionInit     (frameNumber, frameStack);    break;
			case 9:  myTracker.trackToNextFrame_multiTrack_GlobalMotionInit     (frameNumber, frameStack);    break;
			case 10: myTracker.trackToNextFrame_multiTrack_GlobalMotionInit     (frameNumber, frameStack);    break;
     		default: fprintf(stderr,"Un-implemented tracker ID.");
     		         return(-1);
     	}
     	
     	fprintf(stderr,"\rDone tracking through frame %d (Processing framerate: %f)", frameNumber, 1.0/toc());
     	
     	//Render Current Features and check for stop condition (space bar)
     	cvtColor(newFrame, frameWithFeatures, CV_GRAY2RGB);
     	myTracker.features.renderToImage(frameWithFeatures);
     	imshow("Live Display", frameWithFeatures);
		
		key = waitKey(1);
		if (key == ' ') break;
	}
	
	fprintf(stderr,"\nExiting.\n");
     return(0);
}

void displayHelp(void) {
	fprintf(stderr,"\n\r");
	fprintf(stderr,"************************************     Real-Time Tracker     **********************************\n\r");
	fprintf(stderr,"This program runs one of multiple feature trackers in real-time on streaming video. Certain\n\r");
	fprintf(stderr,"parameters are specified through command-line options. The supported options are:\n\r\n\r");
	fprintf(stderr,"-t n       Selects the tracker to be used. Supported values for n are:\n\r");
	fprintf(stderr,"           0:  KLT\n\r");
	fprintf(stderr,"           1:  Unsupported in this release\n\r");
	fprintf(stderr,"           2:  1st-order decent\n\r");
	fprintf(stderr,"           3:  Unsupported in this release\n\r");
	fprintf(stderr,"           4:  Unsupported in this release\n\r");
	fprintf(stderr,"           5:  Multi-tracker with rank penalty - Emp Dim  - uncentered M\n\r");
	fprintf(stderr,"           6:  Multi-tracker with rank penalty - Emp Dim  - centered M\n\r");
	fprintf(stderr,"           7:  Multi-tracker with rank penalty - Nuc Norm - uncentered M\n\r");
	fprintf(stderr,"           8:  Multi-tracker with rank penalty - Nuc Norm - centered M\n\r");
	fprintf(stderr,"           9:  Multi-tracker with rank penalty - Exp Fact - uncentered M\n\r");
	fprintf(stderr,"           10: Multi-tracker with rank penalty - Exp Fact - centered M\n\r\n\r");
	fprintf(stderr,"-v n       Selects the video camera (OpenCV index) to use for tracking.\n\r\n\r");
	fprintf(stderr,"-h         Display this help message.\n\r\n\r");
}

void getFrameWithTimestamp(VideoCapture &cap, Mat &frame, struct timeval &RXTime) {
	Mat frameColor;
	cap >> frameColor;
	gettimeofday(&RXTime, NULL);
	cvtColor(frameColor, frame, CV_RGB2GRAY);
}
