//External Includes
#include "opencv2/opencv.hpp"
#include <stdio.h>
#include <unistd.h>
#include <sys/time.h>
#include "boost/filesystem.hpp"

//Internal Includes
#include "Tracker.h"
#include "ImageStack.h"
#include "parameters.h"
#include "Benchmark.h"
#include "Utilities.h"
#include "rawData.h"

//#define _GNU_SOURCE 1
#include <fenv.h>

using namespace cv;
using namespace std;
using namespace Eigen;
using namespace Utilities;

//This program performs tracking on a pre-recorded data set. It can use one of multiple trackers and it can run
//in different modes. Run with -h command-line switch for a list of run-time options.

//Set default option values
#define DEFAULT_TRACKERID 0
#define DEFAULT_MODE 0
#define DEFAULT_GROUNDTRUTH_PATH ""
#define DEFAULT_INITIALFEATURES_PATH ""
#define DEFAULT_REINIT_TOLERANCE 10.0

//Forward Declarations
void displayHelp(void);
void tracker_Init(Tracker &myTracker, int mode, double reinitTolerance, rawDataLoader &dataLoader, string initFeaturesPath, string groundTruthPath);
void tracker_TrackToNextFrame(Tracker &myTracker, int frameNumber, rawDataLoader &dataLoader, int trackerID, bool renderLive);
Mat toGray(Mat image);

//Constants
const double pi = 3.141592653589793;

int main(int argc, char** argv) {
	int trackerID;
	int mode;
	string groundTruthFilePath;
	string initFeaturesFilePath;
	string sourceFolderPath;
	string targetFilePath;
	double reinitTolerance;
	bool renderLive;
	
	//Parse command-line options
	int tflag = 0;
	int mflag = 0;
	int gflag = 0;
	int iflag = 0;
	int eflag = 0;
	int hflag = 0;
	int rflag = 0;
	
	char *tval = NULL;
	char *mval = NULL;
	char *gval = NULL;
	char *ival = NULL;
	char *eval = NULL;

	int c;
	opterr = 0; //Do not print an error when unrecognized arguments are encountered

	while ((c = getopt (argc, argv, "t:m:g:i:e:hr")) != -1)
		switch (c) {
			case 't':
				tflag = 1;
				tval = optarg;
				break;
			case 'm':
				mflag = 1;
				mval = optarg;
				break;
			case 'g':
				gflag = 1;
				gval = optarg;
				break;
			case 'i':
				iflag = 1;
				ival = optarg;
				break;
			case 'e':
				eflag = 1;
				eval = optarg;
				break;
			case 'h':
				hflag = 1;
				break;
			case 'r':
				rflag = 1;
				break;
			case '?':
				fprintf (stderr, "Unrecognized Option: -%c found. Aborting.\n\r", optopt);
				return(-1);
			default:
				fprintf(stderr, "Unexpected value returned by getopt. Aborting.\n\r");
				return(-1);
	}
	
	if (tflag == 1)
		trackerID = atoi(tval);
	else
		trackerID = DEFAULT_TRACKERID;

	if (mflag == 1)
		mode = atoi(mval);
	else
		mode = DEFAULT_MODE;

	if (gflag == 1)
		groundTruthFilePath = string(gval);
	else
		groundTruthFilePath = string(DEFAULT_GROUNDTRUTH_PATH);
	
	if (iflag == 1)
		initFeaturesFilePath = string(ival);
	else
		initFeaturesFilePath = string(DEFAULT_INITIALFEATURES_PATH);
	
	if (eflag == 1)
		reinitTolerance = atof(eval);
	else
		reinitTolerance = DEFAULT_REINIT_TOLERANCE;
	
	//If help flag was passed, print help text and exit imediately
	if (hflag == 1) {
		displayHelp();
		return(0);
	}
	
	//Process -r option
	if (rflag == 1)
		renderLive = true;
	else
		renderLive = false;
	
	//Check for and process required arguments
	if (argc - optind != 2) {
		fprintf(stderr,"%s has 2 mandatory arguments. Run %s -h for complete usage.\n\r", argv[0], argv[0]);
		return(-1);
	}
	sourceFolderPath = string(argv[optind]);
	targetFilePath = string(argv[optind+1]);
	
     //Done parsing parameters - Try to initialize a raw data loader for the provided data folder and make sure it has valid data
     rawDataLoader dataLoader;
     if (dataLoader.init(sourceFolderPath) != 0) {
     	fprintf(stderr,"Error initializing raw data loader for provided data folder. Aborting.\n\r");
     	return(-1);
     }

	//If we make it this far, we should be ready to track.
	int numFrames = dataLoader.getNumberOfFramesAvailable();
	Tracker myTracker;
	
	//Build Multi-Feature tracker options struct and set up appropriately
     MultiTrackerParameters myMultiFeatureTrackerParams;
     myMultiFeatureTrackerParams.L = 10;
     switch(trackerID) {
     	case 5:  myMultiFeatureTrackerParams.penaltyForm = MultiTrackerParameters::penaltyType::EmpiricalDimension;
     	         myMultiFeatureTrackerParams.penaltyCoefficient = 0.1;     //Videos lightly degraded (new videos)
     	         //myMultiFeatureTrackerParams.penaltyCoefficient = 0.5;     //Old Videos
     	         myMultiFeatureTrackerParams.centerHistoryMatrix = false;
     	         break;
     	case 6:  myMultiFeatureTrackerParams.penaltyForm = MultiTrackerParameters::penaltyType::EmpiricalDimension;
     	         myMultiFeatureTrackerParams.penaltyCoefficient = 0.15;    //Videos lightly degraded (new videos)
     	         //myMultiFeatureTrackerParams.penaltyCoefficient = 0.4;    //Old Videos
     	         myMultiFeatureTrackerParams.centerHistoryMatrix = true;
     	         break;
     	case 7:  myMultiFeatureTrackerParams.penaltyForm = MultiTrackerParameters::penaltyType::NuclearNorm;
     	         myMultiFeatureTrackerParams.penaltyCoefficient = 0.001;   //Videos lightly degraded (new videos)
     	         //myMultiFeatureTrackerParams.penaltyCoefficient = 0.004;   //Old Videos
     	         myMultiFeatureTrackerParams.centerHistoryMatrix = false;
     	         break;
     	case 8:  myMultiFeatureTrackerParams.penaltyForm = MultiTrackerParameters::penaltyType::NuclearNorm;
     	         myMultiFeatureTrackerParams.penaltyCoefficient = 0.0005;   //Videos lightly degraded (new videos)
     	         //myMultiFeatureTrackerParams.penaltyCoefficient = 0.002;   //Old Videos
     	         myMultiFeatureTrackerParams.centerHistoryMatrix = true;
     	         break;
     	case 9:  myMultiFeatureTrackerParams.penaltyForm = MultiTrackerParameters::penaltyType::ExplicitFactorization;
     	         myMultiFeatureTrackerParams.penaltyCoefficient = 0.0015;   //Videos lightly degraded (new videos)
     	         //myMultiFeatureTrackerParams.penaltyCoefficient = 0.05;   //Old Videos
     	         myMultiFeatureTrackerParams.centerHistoryMatrix = false;
     	         break;
     	case 10: myMultiFeatureTrackerParams.penaltyForm = MultiTrackerParameters::penaltyType::ExplicitFactorization;
     	         myMultiFeatureTrackerParams.penaltyCoefficient = 0.002;   //Videos lightly degraded (new videos)
     	         //myMultiFeatureTrackerParams.penaltyCoefficient = 0.05;   //Old Videos
     	         myMultiFeatureTrackerParams.centerHistoryMatrix = true;
     	         break;
     }
	
	//Initialize Tracker
	fprintf(stderr,"Initializing tracker... ");
	tracker_Init(myTracker, mode, reinitTolerance, dataLoader, initFeaturesFilePath, groundTruthFilePath);
	myTracker.multiTrackerParametersPtr = &myMultiFeatureTrackerParams;
	fprintf(stderr,"Done\n\r");
	
	//Track through all available frames
	fprintf(stderr,"Beginning to track\n\r");
	for (int frameNumber = 1; frameNumber < numFrames; frameNumber++) {
		tracker_TrackToNextFrame(myTracker, frameNumber, dataLoader, trackerID, renderLive);
		fprintf(stderr,"\rTracking to frame %d", frameNumber);
	}
	fprintf(stderr,"\n\rTracking Complete.\n\r");
	
	//We are done tracking. Move all current features to dead-tracks feature set
	list< Feature >::iterator featureIter = myTracker.features.Features.begin();
	while (featureIter != myTracker.features.Features.end()) {
		myTracker.deadFeatures.addFeature(*featureIter);
		myTracker.features.removeFeature(featureIter);
	}
	
	if (mode == 2) {
		//Render formated tracker results to stdout. The head and tail tokens are used by other programs to extract the formatted results section. These tokens
		//should not be changed.
		int featureFrameCounter = 0;
		for (list<Feature>::iterator featureIter = myTracker.deadFeatures.Features.begin(); featureIter != myTracker.deadFeatures.Features.end(); featureIter++)
			featureFrameCounter += (int)(featureIter->History.size());
		
		const char * headToken = "******************   Begin Results   ******************\n";
		const char * tailToken = "*******************   End Results   *******************\n";
		fprintf(stdout, "%s", headToken);
		fprintf(stdout, "Total Re-initializations:             %d\n", myTracker.reinitializationCounter);
		fprintf(stdout, "Total feature frame count:            %d\n", featureFrameCounter);
		fprintf(stdout, "Average frames per re-initialization: %f\n", ((double)featureFrameCounter)/((double)(myTracker.reinitializationCounter)));
		fprintf(stdout, "Re-initialization tolerance (pixels): %f\n", reinitTolerance);
		fprintf(stdout, "%s", tailToken);
	}
	
	//Save tracker output
	if (saveTrajectories(myTracker.deadFeatures, targetFilePath) != 0)
		fprintf(stderr,"Error saving final feature trajectories.\n\r");
	
	return(0);
}

void displayHelp(void) {
	fprintf(stderr,"\n\r");
	fprintf(stderr,"**********************************     Post-Processing Tracker     *********************************\n\r");
	fprintf(stderr,"This program runs one of multiple feature trackers on a pre-recorded video. This program supports\n\r");
	fprintf(stderr,"gyro-integrated trackers, provided the data set you try to run on contains inter-frame rotation\n\r");
	fprintf(stderr,"data. This program has 2 mandatory arguments. The first is the absolute path of the source video\n\r");
	fprintf(stderr,"folder on which feature tracking will be run. The second is the absolute path of the output file\n\r");
	fprintf(stderr,"where tracker results will be saved. Other parameters are specified through command-line options.\n\r");
	fprintf(stderr,"The supported options are:\n\r\n\r");
	fprintf(stderr,"-t n       Selects the tracker to be used. Supported values for n are:\n\r");
	fprintf(stderr,"           0:  KLT\n\r");
	fprintf(stderr,"           1:  Unsupported in this release\n\r");
	fprintf(stderr,"           2:  1st-order decent\n\r");
	fprintf(stderr,"           3:  Unsupported in this release\n\r");
	fprintf(stderr,"           4:  Unsupported in this release\n\r");
	fprintf(stderr,"           5:  Multi-tracker with rank penalty - Emp Dim  - uncentered M\n\r");
	fprintf(stderr,"           6:  Multi-tracker with rank penalty - Emp Dim  - centered M\n\r");
	fprintf(stderr,"           7:  Multi-tracker with rank penalty - Nuc Norm - uncentered M\n\r");
	fprintf(stderr,"           8:  Multi-tracker with rank penalty - Nuc Norm - centered M\n\r");
	fprintf(stderr,"           9:  Multi-tracker with rank penalty - Exp Fact - uncentered M\n\r");
	fprintf(stderr,"           10: Multi-tracker with rank penalty - Exp Fact - centered M\n\r\n\r");
	fprintf(stderr,"-m n       Selects the tracker feature management mode. Supported values for n are:\n\r");
	fprintf(stderr,"           0: Detect and manage features autonomously. Replace features that leave field of view\n\r");
	fprintf(stderr,"           1: Load initial feature locations from a file. Do not replace features that leave field\n\r");
	fprintf(stderr,"              of view. If this mode is selected, then an initial features file must be specified\n\r");
	fprintf(stderr,"              through the -i option.\n\r");
	fprintf(stderr,"           2: Load ground truth trajectories from file. When a feature leaves the field of view pre-\n\r");
	fprintf(stderr,"              maturely, or wanders too far from ground truth, the feature is re-initialized to the\n\r");
	fprintf(stderr,"              ground truth location. The number of re-initializations can be used as a performance\n\r");
	fprintf(stderr,"              metric for the tracker.\n\r\n\r");
	fprintf(stderr,"-g path    Sets path for ground truth file (containing ground truth trajectories). This option is\n\r");
	fprintf(stderr,"           needed when the feature management mode is set to 2.\n\r\n\r");
	fprintf(stderr,"-i path    Sets path for initial features file (containing initial locations of features in frame 0)\n\r");
	fprintf(stderr,"           This option is needed when the feature management mode is set to 1.\n\r\n\r");
	fprintf(stderr,"-e tol     Sets allowed amount that a feature can drift from ground truth before being re-\n\r");
	fprintf(stderr,"           initialized when the feature management mode is set to 2.\n\r");
	fprintf(stderr,"           This value is specified in units: pixels\n\r\n\r");
	fprintf(stderr,"-h         Display this help message.\n\r\n\r");
	fprintf(stderr,"-r         Render tracker results to window while tracking.\n\r\n\r");
}

void tracker_Init(Tracker &myTracker, int mode, double reinitTolerance, rawDataLoader &dataLoader, string initFeaturesPath, string groundTruthPath) {
	//Grab frame 0
	Mat frame0;
	dataLoader.getFrame(0, frame0);
	frame0 = toGray(frame0);
	
	ImageStack frameStack;
	frameStack.BuildFromImage(frame0, parameters::M);
	
	switch(mode) {
		case 0:  myTracker.initialize(0, frameStack, mode, string(""));       break;
		case 1:  myTracker.initialize(0, frameStack, mode, initFeaturesPath); break;
		case 2:  myTracker.initialize(0, frameStack, mode, groundTruthPath);  break;
		default: fprintf(stderr,"Unsupported mode. Failed to initialize");
	}
	
	if (mode == 2) {
		//We also need to set the feature re-initialization tolerance prior to tracking
		myTracker.featureReinitTolerance = reinitTolerance;
	}
}

void tracker_TrackToNextFrame(Tracker &myTracker, int frameNumber, rawDataLoader &dataLoader, int trackerID, bool renderLive) {
	//Get new frame of imagery
	Mat frame, grayFrame;
	dataLoader.getFrame(frameNumber, frame);
	grayFrame = toGray(frame);
	
	ImageStack frameStack;
	frameStack.BuildFromImage(grayFrame, parameters::M);
	
	//Step tracker forward to new frame
	switch (trackerID) {
		case 0:  myTracker.trackToNextFrame_KLT                             (frameNumber, frameStack);    break;
		case 1:  fprintf(stderr,"Error: Tracker not supported in this release.\n");                       break;
		case 2:  myTracker.trackToNextFrame_1stOrderDescent_GlobalMotionInit(frameNumber, frameStack);    break;
		case 3:  fprintf(stderr,"Error: Tracker not supported in this release.\n");                       break;
		case 4:  fprintf(stderr,"Error: Tracker not supported in this release.\n");                       break;
		case 5:  myTracker.trackToNextFrame_multiTrack_GlobalMotionInit     (frameNumber, frameStack);    break;
		case 6:  myTracker.trackToNextFrame_multiTrack_GlobalMotionInit     (frameNumber, frameStack);    break;
		case 7:  myTracker.trackToNextFrame_multiTrack_GlobalMotionInit     (frameNumber, frameStack);    break;
		case 8:  myTracker.trackToNextFrame_multiTrack_GlobalMotionInit     (frameNumber, frameStack);    break;
		case 9:  myTracker.trackToNextFrame_multiTrack_GlobalMotionInit     (frameNumber, frameStack);    break;
		case 10: myTracker.trackToNextFrame_multiTrack_GlobalMotionInit     (frameNumber, frameStack);    break;
		default: fprintf(stderr,"Error tracking to frame %d. Tracker ID not recognized.\n\r", frameNumber);
	}
	
	//If requested, render tracker results
	if (renderLive) {
		//We are not going to use "frame" for anything else, so we can write directly to the origional image buffer.
		myTracker.features.renderToImage(frame);
		imshow("Live Display", frame);
		waitKey(1);
	}
}

Mat toGray(Mat image) {
	Mat grayImage;
	if (image.type() == CV_8UC3) {
		cvtColor(image, grayImage, CV_RGB2GRAY);
	}
	else if (image.type() == CV_8UC1){
		grayImage = image;
	}
	else {
		fprintf(stderr,"Function toGray only supports 1 or 3-channel 8-bit images as input.\n\r");
		grayImage = Mat(0,0,CV_8UC1);
	}
	return(grayImage);
}

